class WebScraping():

    """
    WebScraping is a tool that allows you to scrape the web.
    It uses the requests library to scrape the web.

    Examples
    --------

    ```python
    webscraping = WebScraping()
    result = webscraping.scrape("https://www.scrapethissite.com")
    print(result["html"]) # print the html of the page
    print(result["text"]) # print the content of the page merged into a single string
    ```
    """

    def __init__(self, engine:str = "requests", deep:bool = False):

        """
        Initialize the WebScraping tool.

        Parameters:
        ----------
        engine: str, optional
            The engine to use (requests, tavily, selenium. Default is requests)
        deep: bool, optional
            If using tavily, whether to use the advanced extraction mode (default is False)
        """

        if engine == "requests":
            self._engine = _RequestsScraper()
        elif engine == "tavily":
            self._engine = _TavilyScraper(deep=deep)
        elif engine == "selenium":
            self._engine = _SeleniumScraper()
        else:
            raise ValueError(f"Invalid engine: {engine} (must be 'requests', 'tavily', or 'selenium')")

    def scrape(self, url: str):

        """
        Scrape a webpage.

        Parameters:
        ----------
        url: str
            The url to scrape

        Returns:
        -------
        dict
            The response from the scraper.
            html: str
            The html of the page (not available if using tavily)
            text: str
            The content of the page merged into a single string
        """
        response, text_response = self._engine.scrape(url)
        return {"html": response, "text": text_response}
    

from bs4 import BeautifulSoup

class _BaseScraper():

    def __init__(self):
        pass

    def scrape(self, url: str):
        pass

    def _extract_text(self, content: str, remove_whitespace: bool = True) -> str:
        """
        Extract text content from HTML, optionally removing extra whitespace.
        
        Args:
            html_content (str): The HTML content to extract text from
            remove_whitespace (bool): Whether to remove extra whitespace and normalize spaces
            
        Returns:
            str: The extracted text content
        """
        soup = BeautifulSoup(content, 'html.parser')
        
        # Remove script and style elements
        for script in soup(["script", "style"]):
            script.decompose()
        
        # Get text
        text = soup.get_text()
        
        if remove_whitespace:
            # Break into lines and remove leading and trailing space on each
            lines = (line.strip() for line in text.splitlines())
            # Break multi-headlines into a line each
            chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
            # Drop blank lines
            text = ' '.join(chunk for chunk in chunks if chunk)
        
        return text 


import requests

class _RequestsScraper(_BaseScraper):

    def __init__(self):
        super().__init__()

    def scrape(self, url: str):
        response = requests.get(url)
        return response.text, self._extract_text(response.text)
    

from tavily import TavilyClient
from monoai.keys.keys_manager import load_key

class _TavilyScraper(_BaseScraper):

    def __init__(self, deep:bool = False):
        super().__init__()
        load_key("tavily")
        self._client = TavilyClient()
        self._deep = deep

    def scrape(self, url: str):
        response = self._client.extract(url, extract_depth="advanced" if self._deep else "basic")
        response = response["results"][0]
        return None, response["raw_content"]

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import TimeoutException
import time

class _SeleniumScraper(_BaseScraper):
    """
    A scraper that uses Selenium to handle dynamic content and JavaScript.
    This is useful for websites that require JavaScript execution to load content.
    """
    
    def __init__(self, headless: bool = True, wait_time: int = 10):
        """
        Initialize the Selenium scraper.
        
        Args:
            headless (bool): Whether to run Chrome in headless mode (default: True)
            wait_time (int): Maximum time to wait for elements to load (default: 10 seconds)
        """
        super().__init__()
        self._wait_time = wait_time
        self._setup_driver(headless)
    
    def _setup_driver(self, headless: bool):
        """Set up the Chrome WebDriver with specified options."""
        chrome_options = Options()
        if headless:
            chrome_options.add_argument("--headless")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--window-size=1920,1080")
        
        self._driver = webdriver.Chrome(options=chrome_options)
        self._wait = WebDriverWait(self._driver, self._wait_time)
    
    def scrape(self, url: str):
        """
        Scrape a webpage using Selenium.
        
        Args:
            url (str): The URL to scrape
            
        Returns:
            tuple: (html_content, text_content)
        """
        try:
            self._driver.get(url)
            # Wait for the page to load
            time.sleep(2)  # Basic wait for initial page load
            
            # Get the page source after JavaScript execution
            html_content = self._driver.page_source
            text_content = self._extract_text(html_content)
            
            return html_content, text_content
            
        except Exception as e:
            print(f"Error scraping {url}: {str(e)}")
            return None, None
            
    def __del__(self):
        """Clean up the WebDriver when the object is destroyed."""
        if hasattr(self, '_driver'):
            self._driver.quit()



from typing import Dict, Union
from ..prompts.prompt import Prompt
from ..prompts.prompt_chain import PromptChain
from ..prompts.iterative_prompt import IterativePrompt
from litellm import completion

class PromptExecutorMixin:
    """Mixin class to handle prompt execution."""
    
    async def _execute_async(self, prompt: Union[str, Prompt, PromptChain]) -> Dict:
        """
        Execute a prompt asynchronously.
        
        Args:
            prompt: The prompt to execute (string, Prompt, or PromptChain)
            agent: The agent to use for execution
            
        Returns:
            Dictionary containing the response
        """
                
        if isinstance(prompt, PromptChain):
            return await self._execute_chain_async(prompt)
        return self._completion(prompt)
    
    def _execute(self, prompt: Union[str, Prompt, PromptChain]) -> Dict:
        """
        Execute a prompt synchronously.
        
        Args:
            prompt: The prompt to execute (string, Prompt, or PromptChain)
            agent: The agent to use for execution
            
        Returns:
            Dictionary containing the response
        """

        if self._rag:
            response = self._rag.query(prompt)
            print(response)
            if len(response["documents"])>0:
                documents = response.get('documents', [])
                documents = [s for doc_list in documents for s in doc_list]
                documents = '\n'.join(documents)
                prompt += "\nUse also the following information to answer the question: " + documents

        if isinstance(prompt, PromptChain):
            return self._execute_chain(prompt)
        elif isinstance(prompt, IterativePrompt):
            return self._execute_iterative(prompt)
        elif isinstance(prompt, Prompt):
            return self._completion(str(prompt), response_type=prompt.response_type)
        else:
            return self._completion(prompt)

    async def _execute_chain_async(self, chain: PromptChain) -> Dict:
        """
        Execute a prompt chain asynchronously.
        
        Args:
            chain: The prompt chain to execute
            agent: The agent to use for execution
            
        Returns:
            Dictionary containing the final response
        """
        response = None
        for i in range(chain._size):
            current_prompt = chain._format(i, response.output if response else None)
            response = self._completion(current_prompt)
        return response

    def _execute_chain(self, chain: PromptChain) -> Dict:
        """
        Execute a prompt chain synchronously.
        
        Args:
            chain: The prompt chain to execute
            agent: The agent to use for execution
            
        Returns:
            Dictionary containing the final response
        """
        response = None
        for i in range(chain._size):
            current_prompt = chain._format(i, response["choices"][0]["message"]["content"] if response else None)
            response = self._completion(current_prompt)
        return response
    
    def _execute_iterative(self, prompt: IterativePrompt) -> Dict:
        """
        Execute an iterative prompt synchronously.
        
        Args:
            prompt: The iterative prompt to execute
            agent: The agent to use for execution
            
        Returns:
            Dictionary containing the final response
        """
        response = ""
        memory = ""
        for i in range(prompt._size):
            if i > 0 and prompt._has_memory:
                if prompt._retain_all:
                    memory += current_response
                else:
                    memory = current_response   
                current_prompt = prompt._format(i, memory)
            else:
                current_prompt = prompt._format(i)
            current_response = self._completion(current_prompt)

            response += current_response
        return response

    def _completion(self, prompt: str, response_type: str = None) -> Dict:
        
        from pydantic import BaseModel

        class Response(BaseModel):
            response: response_type

        if response_type!=None:
            response_type = Response
        
        messages = [{ "content": prompt,"role": "user"}]
        return completion(model=self.provider+"/"+self.model, 
                          messages=messages, 
                          response_format=response_type,
                          max_tokens=self._max_tokens)



from monoai.chat.base_history import BaseHistory
import os
import json
import uuid
import sqlite3
from monoai.models import Model

class BaseHistory:

    def __init__(self, 
                 history_path: str, 
                 last_n: int=None): 
        self._history_path = history_path
        self._last_n = last_n
        
    def generate_chat_id(self):
        return str(uuid.uuid4())

    def load(self):
        pass

    def save(self):
        pass

    def clear(self):
        pass

class JSONHistory(BaseHistory):
    
    def __init__(self, 
                 history_path: str="histories/", 
                 last_n: int=None): 
        self._history_path = history_path
        self._last_n = last_n
        if not os.path.exists(self._history_path):
            os.makedirs(self._history_path)

    def load(self, chat_id: str):
        with open(self._history_path+chat_id+".json", "r") as f:
            self.messages = json.load(f)
        if self._last_n is not None and len(self.messages) > (self._last_n+1)*2:
            self.messages = [self.messages[0]]+self.messages[-self._last_n*2:]
        return self.messages
    
    def new(self, system_prompt: str):
        chat_id = self.generate_chat_id()
        self.store(chat_id, [{"role": "system", "content": system_prompt}])
        return chat_id

    def store(self, chat_id: str, messages: list):
        with open(self._history_path+chat_id+".json", "w") as f:
            json.dump(messages, f, indent=4)

class SQLiteHistory(BaseHistory):
    
    def __init__(self, db_path: str="histories/chat.db", last_n: int=None):
        self._db_path = db_path
        self._last_n = last_n
        self._init_db()
    
    def _init_db(self):
        os.makedirs(os.path.dirname(self._db_path), exist_ok=True)
        with sqlite3.connect(self._db_path) as conn:
            conn.execute("""
                CREATE TABLE IF NOT EXISTS messages (
                    chat_id TEXT,
                    order_index INTEGER,
                    role TEXT,
                    content TEXT,
                    PRIMARY KEY (chat_id, order_index)
                )
            """)
    
    def load(self, chat_id: str):
        with sqlite3.connect(self._db_path) as conn:
            if self._last_n is not None:
                # Get system message
                cursor = conn.execute(
                    "SELECT role, content FROM messages WHERE chat_id = ? AND order_index = 0",
                    (chat_id,)
                )
                system_message = cursor.fetchone()
                
                # Get last N messages
                cursor = conn.execute(
                    """
                    SELECT role, content 
                    FROM messages 
                    WHERE chat_id = ? 
                    ORDER BY order_index DESC 
                    LIMIT ?
                    """,
                    (chat_id, self._last_n * 2)
                )
                last_messages = [{"role": role, "content": content} for role, content in cursor]
                last_messages.reverse()  # Reverse to get correct order
                
                # Combine system message with last N messages
                self.messages = [{"role": system_message[0], "content": system_message[1]}] + last_messages
            else:
                cursor = conn.execute(
                    "SELECT role, content FROM messages WHERE chat_id = ? ORDER BY order_index",
                    (chat_id,)
                )
                self.messages = [{"role": role, "content": content} for role, content in cursor]
        return self.messages
    
    def new(self, system_prompt: str):
        chat_id = self.generate_chat_id()
        self.store(chat_id, [{"role": "system", "content": system_prompt, "order_index": 0}])
        return chat_id

    def store(self, chat_id: str, messages: list):
        with sqlite3.connect(self._db_path) as conn:
            # Get the last order_index
            cursor = conn.execute(
                "SELECT MAX(order_index) FROM messages WHERE chat_id = ?",
                (chat_id,)
            )
            last_index = cursor.fetchone()[0]
            
            # If no messages exist yet, start from -1
            if last_index is None:
                last_index = -1
            
            # Insert the last two messages with incremented order_index
            for i, message in enumerate(messages[-2:], start=last_index + 1):
                conn.execute(
                    "INSERT INTO messages (chat_id, order_index, role, content) VALUES (?, ?, ?, ?)",
                    (chat_id, i, message["role"], message["content"])
                )
                conn.commit()
        

class HistorySummarizer():

    def __init__(self, model: Model, max_tokens: int=None):
        self._model = model
        self._max_tokens = max_tokens

    def summarize(self, messages: list):
        response = self._model.ask("Summarize the following conversation: "+json.dumps(messages))
        response = response["response"]
        return response




"""
Models are the core of Monoai. They are responsible for executing prompts and returning responses.
"""

from .model import Model
from .multi_model import MultiModel
from .collaborative_model import CollaborativeModel
from .image_model import ImageModel

__all__ = ['Model', 'MultiModel', 'CollaborativeModel', 'ImageModel'] 


import json
from typing import Dict
from ..tokens.token_counter import TokenCounter
from ..tokens.token_cost import TokenCost
from ..prompts.prompt import Prompt

class ResponseProcessorMixin:
    def _process_response(
        self,
        prompt: Prompt,
        response: Dict
    ) -> Dict:
        """
        Process the response and add optional token and cost information.
        
        Args:
            question: The input question
            answer: The model's answer
            provider: Name of the provider
            model: Name of the model
            count_tokens: Whether to count tokens
            count_cost: Whether to calculate costs
            
        Returns:
            Dictionary containing the response and optional stats
        """

        response = response["choices"][0]["message"]["content"]
        
        if isinstance(prompt, Prompt):
            if prompt.response_type != None:
                response = json.loads(response)["response"]

        processed_response = {
            "prompt": str(prompt), 
            "response": response,
            #"messages_trace": response.all_messages(),
            "model": {
                "provider": self.provider, 
                "name": self.model
            }
        }
                        
        return processed_response 


from ..models import Model

class Agent(Model):

    def __init__(self, provider: str, model: str, tools):
        super().__init__(provider, model)
        for tool in tools:
            self._agent._register_tool(tool)

    def run(self, prompt: str):
        return super().ask(prompt)



import uuid

class BaseHistory:

    def generate_chat_id(self):
        return str(uuid.uuid4())

    def load(self):
        pass

    def save(self):
        pass

    def clear(self):
        pass


class WebSearch():

    """
    WebSearch is a tool that allows you to search the web. 
    It uses the DuckDuckGo or Tavily API to search the web.

    Examples
    --------

    ```python
    websearch = WebSearch(engine="duckduckgo", max_results=5)
    result = websearch.search("What is the capital of France?")
    print(result["text"]) # print the result merged into a single string
    print(result["data"]) # print the result as a list of dictionaries
    ```
    """

    def __init__(self, engine: str = "duckduckgo", max_results: int = 5, exclude_domains: list[str] = None):

        """
        Initialize the WebSearch tool.

        Parameters:
        ----------
        engine: str, optional
            The search engine to use (duckduckgo or tavily, default is duckduckgo)
        max_results: int, optional
            The maximum number of results to return (default is 5)
        exclude_domains: list[str], optional
            The domains to exclude from the search (default is None)
        """

        if engine == "duckduckgo":
            self._engine = _DuckDuckGoSearch(max_results, exclude_domains)
        elif engine == "tavily":
            self._engine = _TavilySearch(max_results, exclude_domains)
        else:
            raise ValueError(f"Invalid engine: {engine} (must be 'duckduckgo' or 'tavily')")

    def search(self, query: str):

        """
        Search the web.

        Parameters:
        ----------
        query: str
            The query to search for

        Returns:
        -------
        dict
            The text response from the search engine.
        """
        response, text_response = self._engine.search(query)
        return {"data": response, "text": text_response}
    
from duckduckgo_search import DDGS

class _BaseSearch():

    def __init__(self, max_results: int = 5, exclude_domains: list[str] = None):
        self._max_results = max_results
        self._exclude_domains = exclude_domains

    def search(self, query: str):
        pass

    def _post_process(self, response: list[dict], title_key: str = "title", text_key: str = "body", url_key: str = "url"):
        response = [{"title": item[title_key], "text": item[text_key], "url": item[url_key]} for item in response]
        text_response = "\n\n".join([item["title"] + "\n\n" + item["text"] for item in response])
        return response, text_response

class _DuckDuckGoSearch(_BaseSearch):

    def __init__(self, max_results: int = 5, exclude_domains: list[str] = None):
        super().__init__(max_results, exclude_domains)
        self._client = DDGS()

    def search(self, query: str):
        response = self._client.text(query, max_results=self._max_results)
        return self._post_process(response, title_key="title", text_key="body", url_key="href")
    

from tavily import TavilyClient
from monoai.keys.keys_manager import load_key

class _TavilySearch(_BaseSearch):

    def __init__(self, max_results: int = 5, exclude_domains: list[str] = None):
        super().__init__(max_results, exclude_domains)
        load_key("tavily")
        self._client = TavilyClient()

    def search(self, query: str):
        response = self._client.search(query, max_results=self._max_results, exclude_domains=self._exclude_domains)
        response = response["results"]
        return self._post_process(response, title_key="title", text_key="content", url_key="url")



from litellm import completion, acompletion
from monoai.keys.keys_manager import load_key
from monoai.chat.history import JSONHistory, SQLiteHistory, BaseHistory, HistorySummarizer
from monoai.models import Model

class Chat():
    """
    Chat class is responsible for handling the chat interface and messages history.
    
    Examples
    --------
    Basic usage:
    ```
    chat = Chat(provider="openai", model="gpt-4o-mini")
    response = chat.ask("2+2") # 4
    response = chat.ask("+2") # 6
    ```    

    With history:
    ```

    # Create a new chat
    chat = Chat(provider="openai", model="gpt-4o-mini", history_type="json")
    print(chat.chat_id) # 8cc2bfa3-e9a0-4b82-b46e-3376cd220dd3
    response = chat.ask("Hello! I'm Giuseppe") # Hello!

    # Load a chat
    chat = Chat(provider="openai", model="gpt-4o-mini", history_type="json", chat_id="8cc2bfa3-e9a0-4b82-b46e-3376cd220dd3")
    response = chat.ask("What's my name?") # Your name is Giuseppe
    ```

    With history summarizer:

    ```
    chat = Chat(provider="openai", 
                model="gpt-4o-mini", 
                history_type="json", 
                history_summarizer_provider="openai", 
                history_summarizer_model="gpt-4o-mini", 
                history_summarizer_max_tokens=100)
                
    response = chat.ask("Hello! I'm Giuseppe") # Hello!
    response = chat.ask("What's my name?") # Your name is Giuseppe
    ```
    """

    _HISTORY_MAP = {
        "json": JSONHistory,
        "sqlite": SQLiteHistory
    }

    def __init__(self, 
                 provider: str, 
                 model: str, 
                 system_prompt: str = "You are a helpful assistant.",
                 max_tokens: int = None,
                 history_type: str | BaseHistory = "json", 
                 history_summarizer_provider: str = None, 
                 history_summarizer_model: str = None,
                 history_summarizer_max_tokens: int = None,
                 chat_id: str = None):

        """
        Initialize a new Chat instance.

        Parameters
        ----------
        provider : str
            Name of the provider (e.g., 'openai', 'anthropic')
        model : str
            Name of the model (e.g., 'gpt-4', 'claude-3')
        system_prompt : str | Sequence[str], optional
            System prompt or sequence of prompts
        max_tokens : int, optional
            Maximum number of tokens for each request
        history_type : str | BaseHistory, optional
            The type of history to use for the chat.
        history_summarizer_provider : str, optional
            The provider of the history summarizer.
        history_summarizer_model : str, optional
            The model of the history summarizer.
        history_summarizer_max_tokens : int, optional
            The maximum number of tokens for the history summarizer.
        chat_id : str, optional
            The id of the chat to load, if not provided a new chat will be created
        """

        self._model = model
        self._max_tokens = max_tokens
        load_key(provider)
        self._history_summarizer = None

        if isinstance(history_type, str):
            self._history = self._HISTORY_MAP[history_type]()
        else:
            self._history = history_type

        if history_summarizer_provider is not None and history_summarizer_model is not None:
            self._history_summarizer = HistorySummarizer(Model(provider=history_summarizer_provider, 
                                                               model=history_summarizer_model,
                                                               max_tokens=history_summarizer_max_tokens))

        if chat_id is None:
            self.chat_id = self._history.new(system_prompt)
        else:
            self.chat_id = chat_id

    def ask(self, prompt: str, return_history: bool = False) -> str:
        
        """
        Ask the model a question.

        Parameters
        ----------
        prompt : str
            The question to ask the model
        return_history : bool, optional
            Whether to return the full history of the chat or only the response

        Returns
        -------
        str
            The response from the model
        list
            The full history of the chat if return_history is True
        """

        messages = self._history.load(self.chat_id)

        if self._history_summarizer is not None:
            summarized = self._history_summarizer.summarize(messages)
            messages.append({"role": "user", "content": prompt})
            ask_messages = [messages[0], messages[-1]]
            print(summarized)
            print(ask_messages[0])
            ask_messages[0]["content"] += "\n\nHere is the summary of the conversation so far:\n\n" + summarized
        else:
            messages.append({"role": "user", "content": prompt})
            ask_messages = messages

        response = completion(
            model=self._model,
            messages=ask_messages,
            max_tokens=self._max_tokens
        )
        
        response = response["choices"][0]["message"]["content"]
        messages.append({"role": "assistant", "content": response})
        
        self._history.store(self.chat_id, messages)
        if return_history:
            return messages
        else:
            return response
    
    async def ask_async(self, prompt: str):

        """
        Ask the model a question and stream the response.

        Parameters
        ----------
        prompt : str
            The question to ask the model

        Returns
        -------
        AsyncGenerator[str, None]
            A generator that yields the response from the model
        """
        messages = self._history.load(self.chat_id)
        messages.append({"role": "user", "content": prompt})

        response = await acompletion(
            model=self._model,
            messages=messages,
            stream=True,
            max_tokens=self._max_tokens
        )

        response_text = ""
        async for part in response:
            part = part["choices"][0]["delta"]["content"] or ""
            response_text += part
            yield part
        messages.append({"role": "assistant", "content": response_text})
        self._history.store(self.chat_id, messages)



